#define _XOPEN_SOURCE 600
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <pty.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#define TRIES 20

void
do_write(int fd, char const *str)
{
    int r = write(fd, str, strlen(str));
    printf("write(%d, %s, %d) = %d\n", fd, str, (int) strlen(str), r);
}

int
do_read(int fd)
{
    char buf[2048];
    int r = read(fd, buf, 2048);
    printf("read(%d, buf, 2048) = %d\n", fd, r);
    if (r > 0) {
        printf("buf = |%.*s|\n", r, buf);
    }
    return r;
}

void
sleep100ms()
{
    struct timespec sleeptime;

    /* Delay for 0.1 seconds */
    sleeptime.tv_sec = 0;
    sleeptime.tv_nsec = 1000000000 / 10;

    nanosleep(&sleeptime, NULL);
}

int main(int argc, char **argv[])
{
    int r;
    int i;
    int mfd;
    int found = 0;
    char const *pts;
    int sfd;
    int disc = N_HDLC;
    struct timespec sleeptime;

    mfd = posix_openpt(O_RDWR|O_NOCTTY);
    printf("posix_openpt = %d\n", mfd);

    r = grantpt(mfd);
    printf("grantpt(%d)  = %d\n", mfd, r);

    r = unlockpt(mfd);
    printf("unlockpt(%d) = %d\n", mfd, r);

    pts = ptsname(mfd);
    printf("ptsname(%d) = %s\n", mfd, pts);

    sfd = open(pts, O_RDWR|O_NOCTTY|O_NONBLOCK);
    printf("open(%s) = %d\n", pts, sfd);

    r = ioctl(mfd, TIOCSETD, &disc);
    printf("ioctl(%d, TIOCSETD) = %d\n", mfd, r);

    r = ioctl(sfd, TIOCSETD, &disc);
    printf("ioctl(%d, TIOCSETD) = %d\n", sfd, r);

    /* Try up to TRIES times for the writes to be combined in a single read */
    for (i=1; i<=TRIES; i++) {
        do_write(mfd, "Hello ");
        if (argc > 1) {
            printf("Sleeping for 100ms between writes\n");
            sleep100ms();
        }
        do_write(mfd, "world");
        sleep100ms();
        do {
            r = do_read(sfd);
            if (r == 11) {
                printf("HAHA!  Try #%d: The two writes were combined in a single read!\n", i);
                found = 1;
            }
        } while (r > 0 && !found);
        if (found) {
            break;
        }
    }

    if (!found) {
        printf("Hmm, after %d tries, never saw the two writes combined in a single read.\n", TRIES);
    }
    close(sfd);
    close(mfd);
    exit(0);
}

